{-# LANGUAGE ScopedTypeVariables #-}

module Tests where

import Test.Tasty
import Test.Tasty.HUnit

import Data.List (sort)
import Data.Maybe (fromMaybe)

import LeftistHeap
import Huffman

run = defaultMain $ testGroup "All tests" [leftistHeapTests, huffmanTests]

-- Used to test if ONLY the precomputed rank is being accessed.
-- Accessing any other field results in an error.
forkPrecomputedRank4 :: LeftistHeap Char
forkPrecomputedRank4 = Fork 4 
    (error "value of forkPrecomputedRank4 was evaluated")
    (error "left subtree of forkPrecomputedRank4 was evaluated")
    (error "right subtree of forkPrecomputedRank4 was evaluated")

fork1Elem :: LeftistHeap Char
fork1Elem = Fork 1 'y' Leaf Leaf

fork2Elem :: LeftistHeap Char
fork2Elem = Fork 1 'x' (Fork 1 'y' Leaf Leaf) Leaf

-- result of merging fork1Elem with fork2Elem
forks1And2Elem :: LeftistHeap Char
forks1And2Elem = Fork 2 'x'
                    (Fork 1 'y' Leaf Leaf) 
                    (Fork 1 'y' Leaf Leaf)

-- Used to test if ONLY the value at the root of the heap is accessed.
-- Accessing any other field results in an error.
presortedHeapMinX :: LeftistHeap Char
presortedHeapMinX = Fork 
    (error "rank of presortedHeapMinX was evaluated")
    'X'
    (error "left subtree of presortedHeapMinX was evaluated")
    (error "right subtree of presortedHeapMinX was evaluated")

leftistHeapTests :: TestTree
leftistHeapTests = testGroup "LeftistHeap"
    [ testCase "rank Leaf == 0" $
        rank (Leaf :: LeftistHeap Char) @?= 0
    , testCase "rank (Fork 4 undefined undefined undefined) == 4" $
        rank forkPrecomputedRank4 @?= 4
    , testCase "makeFork: rank of right subtree + 1" $
        makeFork 'a' fork2Elem fork1Elem @?= Fork 2 'a' fork2Elem fork1Elem
    , testCase "makeFork: pick higher ranking subtree as left subtree" $
        makeFork 'a' fork2Elem Leaf @?= makeFork 'a' Leaf fork2Elem
    , testCase "merge fork2Elem fork1Elem == forks1And2Elem" $
        merge fork2Elem fork1Elem @?= forks1And2Elem
    , testCase "insert 'y' empty" $
        insert 'y' empty @?= fork1Elem
    , testCase "root must be the minimum" $
        insert 'y' (insert 'x' empty) @?= fork2Elem
    , testCase "viewMin (Fork undefined 'X' undefined undefined) == Just 'X'" $
        viewMin presortedHeapMinX @?= Just 'X'
    , testCase "extractMin fork2Elem == Just ('x', fork1Elem)" $
        extractMin fork2Elem @?= Just ('x', fork1Elem)
    , testCase "(toList . fromList) [3,5,1,4,2] == [1..5]" $
        (toList . fromList) [3,5,1,4,2] @?= [1..5]
    , testCase "(toList . fromList) [3,3,3] == [3,3,3]" $
        (toList . fromList) [3,3,3] @?= [3,3,3]
    ]

abrakadabraFreqs :: [(Char, Frequency)]
abrakadabraFreqs = [('a',5),('b',2),('d',1),('k',1),('r',2)]

lookupFrequency :: Eq a => a -> [(a, Frequency)] -> Frequency
lookupFrequency symbol freqs = fromMaybe 0 (symbol `lookup` freqs)

-- Given a symbol with its associated codeword as well as symbol frequencies, 
-- totalCodewordLength will return the total number of bits used to encode the
-- symbol.
totalCodewordLength :: Eq a => [(a, Frequency)] -> (a, [Bit]) -> Int
totalCodewordLength freqs (symbol, bits) =
  lookupFrequency symbol freqs * length bits

totalCodeLength :: Ord a => [a] -> Maybe Int
totalCodeLength (symbols :: [a]) = codeLengthFromTree <$> huffmanTree freqs
  where
    codeLengthFromTree :: HuffmanTree a -> Int
    codeLengthFromTree = sum . map (totalCodewordLength freqs) . codewords
    freqs :: [(a, Frequency)]
    freqs = frequencies symbols

encodeThenDecode :: (Show a, Ord a) => [a] -> [a]
encodeThenDecode (symbols :: [a]) = (dec . enc) symbols
  where
    Just tree = (huffmanTree . frequencies) symbols
    dec :: [Bit] -> [a]
    dec bits = case decode bits tree of
        Nothing      -> error $ "dec: Failed decoding bits " ++ show bits
        Just symbols -> symbols
    enc :: [a] -> [Bit]
    enc symbols = case encode symbols (codewords tree) of
        Nothing   -> error $ "enc: Failed encoding symbols " ++ show symbols
        Just bits -> bits

huffmanTests :: TestTree
huffmanTests = testGroup "Huffman"
    [ testCase "frequencies \"abrakadabra\"" $
        sort (frequencies "abrakadabra") @?= abrakadabraFreqs
    , testCase "frequenciesAscii \"abrakadabra\"" $
        sort (frequenciesAscii "abrakadabra") @?= abrakadabraFreqs
    , testCase "minimum total codelength for \"abrakadabra\" is 23" $
        totalCodeLength "abrakadabra" @?= Just 23
    , testCase "encode . decode == id   (for \"abrakadabra\")" $
        encodeThenDecode "abrakadabra" @?= "abrakadabra"
    ]